"""Python SDK for reading/writing to/from pvsite database."""

from .connection import DatabaseConnection
from .sqlmodels import (
    APIRequestSQL,
    ForecastSQL,
    ForecastValueSQL,
    GenerationSQL,
    InverterSQL,
    SiteGroupSQL,
    SiteSQL,
    StatusSQL,
    UserSQL,
)

__version__ = "1.0.28"



""" Utils for GSP and DNO. """

try:
    import pyproj
except ImportError:
    print("You might want to install pyproj")  # noqa

# OSGB is also called "OSGB 1936 / British National Grid -- United
# Kingdom Ordnance Survey".  OSGB is used in many UK electricity
# system maps, and is used by the UK Met Office UKV model.  OSGB is a
# Transverse Mercator projection, using 'easting' and 'northing'
# coordinates which are in meters.  See https://epsg.io/27700

# WGS84 is short for "World Geodetic System 1984", used in GPS. Uses
# latitude and longitude.

OSGB = 27700
WGS84 = 4326
WGS84_CRS = f"EPSG:{WGS84}"


def lat_lon_to_osgb(lat: float, lon: float) -> [float, float]:
    """
    Change lat, lon to a OSGB coordinates.

    lat: latitude
    lon: longitude

    Return: 2-tuple of x (east-west), y (north-south).

    """
    return transformers.lat_lon_to_osgb.transform(lat, lon)


class Transformers:
    """
    Class to store transformation from one Grid to another.

    Its good to make this only once, but need the
    option of updating them, due to out of data grids.
    """

    def __init__(self):
        """Init."""
        self._osgb_to_lat_lon = None
        self._lat_lon_to_osgb = None
        self._osgb_to_geostationary = None
        self.make_transformers()

    def make_transformers(self):
        """
        Make transformers.

         Nice to only make these once, as it makes calling the functions below quicker
        """
        self._lat_lon_to_osgb = pyproj.Transformer.from_crs(crs_from=WGS84, crs_to=OSGB)

    @property
    def lat_lon_to_osgb(self):
        """lat-lon to OSGB property."""
        return self._lat_lon_to_osgb


transformers = Transformers()



"""
Functions for reading from the PVSite database
"""

from .generation import get_pv_generation_by_sites, get_pv_generation_by_user_uuids
from .latest_forecast_values import get_latest_forecast_values_by_site
from .site import (
    get_all_sites,
    get_site_by_client_site_id,
    get_site_by_client_site_name,
    get_site_by_uuid,
    get_sites_by_country,
    get_sites_from_user,
)
from .status import get_latest_status
from .user import (
    get_all_last_api_request,
    get_all_site_groups,
    get_all_users,
    get_api_requests_for_one_user,
    get_site_group_by_name,
    get_user_by_email,
)



""" GSP functions for UK regions. """

import logging
import os

import pandas as pd

from pvsite_datamodel.write.data.utils import lat_lon_to_osgb

try:
    import geopandas as gpd
    from shapely.geometry import Point
except ImportError:
    print("You might want to install geopandas")  # noqa

logger = logging.getLogger(__name__)

dir_path = os.path.dirname(os.path.realpath(__file__))
gsp_local_file = f"{dir_path}/gsp"
gsp_names = pd.read_csv(f"{dir_path}/gsp_new_ids_and_names-edited.csv")


def get_gsp(latitude, longitude) -> dict:
    """
    Get a DNO from latitude and longitude.

    :param latitude:
    :param longitude:

    :return: dno is this format {"dno_id": dno_id, "name": dno_name, "long_name": dno_long_name}=
    """

    # load file
    gsp = gpd.read_file(gsp_local_file)

    # change lat lon to osgb
    x, y = lat_lon_to_osgb(lat=latitude, lon=longitude)
    point = Point(x, y)

    # select gsp
    mask = gsp.contains(point)
    gsp = gsp[mask]

    # format gsp
    if len(gsp) == 1:
        gsp = gsp.iloc[0]
        gsp_details = gsp_names[gsp_names["gsp_name"] == gsp.GSPs]
        gsp_id = gsp_details.index[0]
        gsp_details = gsp_details.iloc[0]
        name = gsp_details["region_name"]

        gsp_dict = {"gsp_id": str(gsp_id), "name": name}
    else:
        gsp_dict = {"gsp_id": "999", "name": "unknown"}

    return gsp_dict



""" The script adds the relevant GSP to the sites.

Might need to install nowcasting_dataset

and we want to added the gsp as {gsp_id}|{gsp_nam} into the database

1. Load in dno data from NG
2. Load all sites
3. For each site add dno

"""

import logging
import os

from pvsite_datamodel.write.data.utils import lat_lon_to_osgb

try:
    import geopandas as gpd
    from shapely.geometry import Point
except ImportError:
    print("You might want to install geopandas")  # noqa

logger = logging.getLogger(__name__)
dir_path = os.path.dirname(os.path.realpath(__file__))
dno_local_file = f"{dir_path}/dno"


def get_dno(latitude, longitude) -> dict:
    """
    Convert a latitude and longitude and returns the dno.

    :param latitude:
    :param longitude:

    :return: dno is this format {"dno_id": dno_id, "name": dno_name, "long_name": dno_long_name}=
    """

    # load file
    dno = gpd.read_file(dno_local_file)

    # change lat lon to osgb
    x, y = lat_lon_to_osgb(lat=latitude, lon=longitude)
    point = Point(x, y)

    # select dno
    mask = dno.contains(point)
    dno = dno[mask]

    # format dno
    if len(dno) == 1:
        dno = dno.iloc[0]

        dno_id = dno["ID"]
        name = dno["Name"]
        long_name = dno["LongName"]

        dno_dict = {"dno_id": str(dno_id), "name": name, "long_name": long_name}
        logger.debug(dno_dict)
    else:
        dno_dict = {"dno_id": "999", "name": "unknown", "long_name": "unknown"}

    return dno_dict


#



""" Functions for reading user data from the database. """

import logging
from datetime import datetime
from typing import List, Optional

from sqlalchemy.orm import Session, contains_eager

from pvsite_datamodel.sqlmodels import APIRequestSQL, SiteGroupSQL, UserSQL

logger = logging.getLogger(__name__)


def get_user_by_email(session: Session, email: str):
    """
    Get user by email. If user does not exist, make one.

    :param session: database session
    :param email: email of user
    :return: user object
    """

    user = session.query(UserSQL).filter(UserSQL.email == email).first()

    if user is None:
        logger.info(f"User with email {email} not found, so making one")

        # making a new site group
        site_group = SiteGroupSQL(site_group_name=f"site_group_for_{email}")
        session.add(site_group)
        session.commit()

        # make a new user
        user = UserSQL(email=email, site_group_uuid=site_group.site_group_uuid)
        session.add(user)
        session.commit()

    return user


# get all users
def get_all_users(session: Session) -> List[UserSQL]:
    """
    Get all users from the database.

    :param session: database session
    """
    query = session.query(UserSQL)

    query = query.order_by(UserSQL.email.asc())

    users = query.all()

    return users


def get_site_group_by_name(session: Session, site_group_name: str):
    """
    Get site group by name. If site group does not exist, make one.

    :param session: database session
    :param site_group_name: name of site group
    :return: site group object
    """

    site_group = (
        session.query(SiteGroupSQL).filter(SiteGroupSQL.site_group_name == site_group_name).first()
    )

    if site_group is None:
        logger.info(f"Site group with name {site_group_name} not found, so making one")

        # make a new site group
        site_group = SiteGroupSQL(site_group_name=site_group_name)
        session.add(site_group)
        session.commit()

    return site_group


# get all site groups
def get_all_site_groups(session: Session) -> List[SiteGroupSQL]:
    """
    Get all site groups from the database.

    :param session: database session
    """
    query = session.query(SiteGroupSQL)

    query = query.order_by(SiteGroupSQL.site_group_name.asc())

    site_groups = query.all()

    return site_groups


def get_all_last_api_request(session: Session) -> List[APIRequestSQL]:
    """
    Get all last api requests for all users.

    :param session: database session
    :return:
    """

    last_requests_sql: [APIRequestSQL] = (
        session.query(APIRequestSQL)
        .distinct(APIRequestSQL.user_uuid)
        .join(UserSQL)
        .options(contains_eager(APIRequestSQL.user))
        .populate_existing()
        .order_by(APIRequestSQL.user_uuid, APIRequestSQL.created_utc.desc())
        .all()
    )

    return last_requests_sql


def get_api_requests_for_one_user(
    session: Session,
    email: str,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
) -> List[APIRequestSQL]:
    """
    Get all api requests for one user.

    :param session: database session
    :param email: user email
    :param start_datetime: only get api requests after start datetime
    :param end_datetime: only get api requests before end datetime
    """

    query = session.query(APIRequestSQL).join(UserSQL).filter(UserSQL.email == email)

    if start_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc >= start_datetime)

    if end_datetime is not None:
        query = query.filter(APIRequestSQL.created_utc <= end_datetime)

    api_requests: [APIRequestSQL] = query.order_by(APIRequestSQL.created_utc.desc()).all()

    return api_requests



"""
Write helpers for the Forecast and ForecastValues table.
"""

import logging

import pandas as pd
from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import ForecastSQL, ForecastValueSQL

_log = logging.getLogger(__name__)


def insert_forecast_values(
    session: Session,
    forecast_meta: dict,
    forecast_values_df: pd.DataFrame,
):
    """Insert a dataframe of forecast values and forecast meta info into the database.

    :param session: sqlalchemy session for interacting with the database
    :param forecast_meta: Meta info about the forecast values
    :param forecast_values_df: dataframe with the data to insert
    """

    forecast = ForecastSQL(**forecast_meta)
    session.add(forecast)

    # Flush to get the Forecast's primary key.
    session.flush()

    rows = forecast_values_df.to_dict("records")
    session.bulk_save_objects(
        [
            ForecastValueSQL(
                **row,
                forecast_uuid=forecast.forecast_uuid,
            )
            for row in rows
        ]
    )
    session.commit()



"""Functions to read from the Status table."""

from typing import Optional

from sqlalchemy.orm import Session

from pvsite_datamodel.sqlmodels import StatusSQL


def get_latest_status(session: Session) -> Optional[StatusSQL]:
    """Get the latest entry in the status table.

    Return None if there are no entries.
    :param session: database session
    :return: Latest StatusSQL object, or None
    """
    # Query status table, ordered by created time
    query = session.query(StatusSQL)
    query = query.order_by(StatusSQL.created_utc.desc())

    # get the top result
    latest_status = query.first()

    return latest_status
